@@ -16,19 +16,27 @@
 
 //! `FEEL` definitions.
 
+extern crate ascii_tree;
 extern crate chrono;
 extern crate chrono_tz;
-pub extern crate dmntk_common;
+#[macro_use]
+extern crate derivative;
+extern crate dmntk_common;
 #[macro_use]
 extern crate lazy_static;
 extern crate regex;
-extern crate serde;
+extern crate rust_decimal;
+extern crate rust_decimal_macros;
 #[macro_use]
-extern crate serde_derive;
+extern crate thiserror;
 
 pub use ast::{AstNode, OptAstNode};
+pub use evaluator::Evaluator;
+pub use function::FunctionBody;
 pub use names::{Name, OptName, QualifiedName};
+pub use numbers::FeelNumber;
 pub use scope::Scope;
+pub use strings::ToFeelString;
 pub use temporal::date::FeelDate;
 pub use temporal::dt_duration::FeelDaysAndTimeDuration;
 pub use temporal::ym_duration::FeelYearsAndMonthsDuration;
@@ -36,31 +44,15 @@ pub use temporal::{subtract, FeelDateTime, FeelTime};
 pub use types::{is_built_in_type_name, FeelType};
 
 mod ast;
+mod ast_tree;
 pub mod bif;
 pub mod context;
-pub mod dto;
+mod evaluator;
+mod function;
 mod names;
+mod numbers;
 mod scope;
+mod strings;
 mod temporal;
 mod types;
 pub mod values;
-
-use std::os::raw::c_double;
-
-// External function for rounding using half even mode.
-extern "C" {
-  fn roundeven(value: c_double) -> c_double;
-}
-
-/// Rounds the nearest half even with specified scale.
-pub fn round_half_to_even(number: f64, scale: f64) -> f64 {
-  let value = (number * 10_f64.powf(scale + 1.0)).trunc() / 10.0;
-  let rounded = unsafe { roundeven(value) };
-  rounded / 10_f64.powf(scale)
-}
-
-/// Checks if the floating-point value is an integer,
-/// that is, has no (or very small) fractional part.
-pub fn is_integer(value: f64) -> bool {
-  value.fract() < f64::EPSILON
-}
