@@ -12,23 +12,32 @@
  * SOFTWARE.
  */
 
-//! FEEL scope.
+//! `FEEL` scope.
 
 use crate::context::FeelContext;
 use crate::values::Value;
 use crate::Name;
-use dmntk_common::{Jsonify, Stringify};
+use dmntk_common::Jsonify;
 use std::cell::RefCell;
 use std::collections::HashSet;
 
+/// Creates a scope.
+#[macro_export]
+macro_rules! scope {
+  () => {{
+    Scope::default()
+  }};
+}
+
 /// The `FEEL` scope.
+#[derive(Debug)]
 pub struct Scope {
   /// The stack of contexts.
   contexts: RefCell<Vec<FeelContext>>,
 }
 
 impl Default for Scope {
-  /// Creates a default [Scope] containing one default [FeelContext].
+  /// Creates a default [Scope] containing single default [FeelContext].
   fn default() -> Self {
     Self {
       contexts: RefCell::new(vec![FeelContext::default()]),
@@ -45,12 +54,13 @@ impl From<FeelContext> for Scope {
   }
 }
 
-impl Stringify for Scope {
-  /// Converts this [Scope] to its `TEXT` representation.
-  fn stringify(&self) -> String {
-    format!(
+impl std::fmt::Display for Scope {
+  /// Converts this [Scope] to its textual representation.
+  fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+    write!(
+      f,
       "[{}]",
-      self.contexts.borrow_mut().iter().map(|ctx| ctx.stringify()).collect::<Vec<String>>().join(",")
+      self.contexts.borrow_mut().iter().map(|ctx| ctx.to_string()).collect::<Vec<String>>().join(", ")
     )
   }
 }
@@ -60,7 +70,7 @@ impl Jsonify for Scope {
   fn jsonify(&self) -> String {
     format!(
       "[{}]",
-      self.contexts.borrow_mut().iter().map(|ctx| ctx.jsonify()).collect::<Vec<String>>().join(",")
+      self.contexts.borrow_mut().iter().map(|ctx| ctx.to_string()).collect::<Vec<String>>().join(", ")
     )
   }
 }
@@ -76,10 +86,16 @@ impl Scope {
   pub fn push(&self, ctx: FeelContext) {
     self.contexts.borrow_mut().push(ctx)
   }
-  /// Takes and returns a context from the top of the scope stack.
+  /// Takes and returns a context from the top of the stack.
   pub fn pop(&self) -> Option<FeelContext> {
     self.contexts.borrow_mut().pop()
   }
+  /// Peeks a to context from the top of the stack.
+  /// If the stack is empty, the default context is returned.
+  pub fn peek(&self) -> FeelContext {
+    //FIXME maybe returning a reference is enough???
+    self.contexts.borrow_mut().last().map_or(FeelContext::default(), |ctx| ctx.clone())
+  }
   /// Returns a vector of flattened keys in all contexts in scope.
   pub fn flatten_keys(&self) -> HashSet<String> {
     self
@@ -109,29 +125,35 @@ impl Scope {
     }
     None
   }
-  /// Sets a value for entry name in [FeelContext] placed on the top of the scope stack (last context).
+  /// Sets a specified value for entry name in [FeelContext] placed on the top of the scope stack (last context).
   pub fn set_entry(&self, name: &Name, value: Value) {
     if let Some(context) = self.contexts.borrow_mut().last_mut() {
       context.set_entry(name, value);
     }
   }
+  /// Sets a null value for entry name in [FeelContext] placed on the top of the scope stack (last context).
+  pub fn insert_null(&self, name: Name) {
+    if let Some(context) = self.contexts.borrow_mut().last_mut() {
+      context.set_null(name);
+    }
+  }
 }
 
 #[cfg(test)]
 mod tests {
   use crate::context::FeelContext;
   use crate::values::Value;
-  use crate::{Name, Scope};
-  use dmntk_common::Stringify;
+  use crate::{scope, value_number, FeelNumber, Name, Scope};
 
   #[test]
   fn test_scope_default() {
-    assert_eq!("[{}]", Scope::default().stringify());
+    assert_eq!("[{}]", Scope::default().to_string());
+    assert_eq!("[{}]", scope!().to_string());
   }
 
   #[test]
   fn test_scope_new() {
-    assert_eq!("[]", Scope::new().stringify());
+    assert_eq!("[]", Scope::new().to_string());
   }
 
   #[test]
@@ -139,9 +161,9 @@ mod tests {
     let scope = Scope::new();
     let ctx = FeelContext::default();
     scope.push(ctx);
-    assert_eq!("[{}]", scope.stringify());
+    assert_eq!("[{}]", scope.to_string());
     let scope: Scope = FeelContext::default().into();
-    assert_eq!("[{}]", scope.stringify());
+    assert_eq!("[{}]", scope.to_string());
   }
 
   #[test]
@@ -150,39 +172,39 @@ mod tests {
     scope.push(FeelContext::default());
     scope.push(FeelContext::default());
     scope.push(FeelContext::default());
-    assert_eq!("[{},{},{}]", scope.stringify());
+    assert_eq!("[{}, {}, {}]", scope.to_string());
     let scope = Scope::default();
     scope.push(FeelContext::default());
     scope.push(FeelContext::default());
     scope.push(FeelContext::default());
-    assert_eq!("[{},{},{},{}]", scope.stringify());
+    assert_eq!("[{}, {}, {}, {}]", scope.to_string());
   }
 
   #[test]
   fn test_scope_single_context() {
     let scope = Scope::default();
-    assert_eq!("[{}]", scope.stringify());
+    assert_eq!("[{}]", scope.to_string());
     let name_a = Name::from("a");
     let name_b = Name::from("b");
-    scope.set_entry(&name_a, Value::Number(49.5));
-    assert_eq!("[{a:49.5}]", scope.stringify());
-    scope.set_entry(&name_b, Value::Number(50.0));
-    assert_eq!("[{a:49.5,b:50}]", scope.stringify());
+    scope.set_entry(&name_a, value_number!(495, 1));
+    assert_eq!("[{a: 49.5}]", scope.to_string());
+    scope.set_entry(&name_b, value_number!(50));
+    assert_eq!("[{a: 49.5, b: 50}]", scope.to_string());
     scope.pop();
-    assert_eq!("[]", scope.stringify());
+    assert_eq!("[]", scope.to_string());
   }
 
   #[test]
   fn test_scope_no_context() {
     let scope = Scope::new();
-    assert_eq!("[]", scope.stringify());
+    assert_eq!("[]", scope.to_string());
     let name_a = Name::from("a");
     let name_b = Name::from("b");
-    scope.set_entry(&name_a, Value::Number(1.25));
-    assert_eq!("[]", scope.stringify());
-    scope.set_entry(&name_b, Value::Number(1.75));
-    assert_eq!("[]", scope.stringify());
+    scope.set_entry(&name_a, value_number!(125, 2));
+    assert_eq!("[]", scope.to_string());
+    scope.set_entry(&name_b, value_number!(175, 2));
+    assert_eq!("[]", scope.to_string());
     scope.pop();
-    assert_eq!("[]", scope.stringify());
+    assert_eq!("[]", scope.to_string());
   }
 }
