@@ -14,30 +14,64 @@
  * SOFTWARE.
  */
 
-//! FEEL values.
+//! `FEEL` values.
 
 use self::errors::*;
 use crate::bif::Bif;
 use crate::context::FeelContext;
 use crate::names::Name;
+use crate::numbers::FeelNumber;
+use crate::strings::ToFeelString;
 use crate::temporal::date::FeelDate;
 use crate::temporal::dt_duration::FeelDaysAndTimeDuration;
 use crate::temporal::ym_duration::FeelYearsAndMonthsDuration;
 use crate::temporal::{FeelDateTime, FeelTime};
 use crate::types::FeelType;
-use crate::AstNode;
-use dmntk_common::{null_with_trace, Feelify, Jsonify, Result, Stringify};
+use crate::FunctionBody;
+use dmntk_common::{Jsonify, Result};
 use std::collections::BTreeMap;
 use std::convert::TryFrom;
 use std::ops::Deref;
 use std::str::FromStr;
 
-/// Utility constant for `null` value of type `Null`.
-pub const VALUE_NULL: Value = Value::Null(None);
+/// Creates `Value::Null` with optional tracing message.
+///
+/// # Examples
+///
+/// ```
+/// use crate::dmntk_feel::{value_null, values::Value};
+///
+/// let v = value_null!();
+/// assert_eq!("null", v.to_string());
+///
+/// let v = value_null!("missing input parameter");
+/// assert_eq!("null(missing input parameter)", v.to_string());
+///
+/// let v = value_null!("integer out of range {}..{}", 1, 100);
+/// assert_eq!("null(integer out of range 1..100)", v.to_string());
+/// ```
+#[macro_export]
+macro_rules! value_null {
+  ($f:expr, $($a:tt)*) => {
+    Value::Null(Some(format!($f, $($a)*)))
+  };
+  ($l:expr) => {
+    Value::Null(Some(format!("{}", $l)))
+  };
+  () => {
+    Value::Null(None)
+  };
+}
 
-/// Utility constant for `null` value of type `Null` with additional tracing message.
-/// Tracing messages may be used for diagnostic purposes.
-pub const VALUE_NULL_WITH_TRACE: Value = Value::Null(None);
+#[macro_export]
+macro_rules! value_number {
+  ($n:expr) => {{
+    Value::Number(FeelNumber::new_int($n))
+  }};
+  ($n:expr, $s:expr) => {
+    Value::Number(FeelNumber::new($n, $s))
+  };
+}
 
 /// Utility constant for value `true `of type `Boolean`.
 pub const VALUE_TRUE: Value = Value::Boolean(true);
@@ -45,23 +79,7 @@ pub const VALUE_TRUE: Value = Value::Boolean(true);
 /// Utility constant for value `false` of type `Boolean`.
 pub const VALUE_FALSE: Value = Value::Boolean(false);
 
-/// Function body may be defined in `FEEL` or `DMN` in many ways,
-/// so this enumeration is the representation of all these cases.
-#[derive(Debug, Clone, PartialEq)]
-pub enum FunctionBody {
-  /// Function body created from model context.
-  // Context(Context),
-  /// Function body created from `FEEL` textual representation.
-  LiteralExpression(AstNode),
-  /// Function body created from decision table.
-  /// DecisionTable(DecisionTable),
-  /// Function body is defined as a call to an externally defined function,
-  /// it may be a Java function or PMML document. The value in this
-  /// representation `SHALL` be a context.
-  External(Box<Value>),
-}
-
-/// ???
+/// `FEEL` value.
 #[derive(Debug, Clone, PartialEq)]
 pub enum Value {
   /// Value representing `FEEL` boolean type.
@@ -70,11 +88,8 @@ pub enum Value {
   /// Value for storing built-in function definition.
   BuiltInFunction(Bif),
 
-  /// Value for storing built-in type definition.
-  BuiltInType(FeelType),
-
-  /// Value for storing a collection of values representing a comma-separated list of values.
-  CommaList(Values),
+  /// Value representing a collection of comma-separated list of expressions.
+  ExpressionList(Values),
 
   /// Value representing a context.
   Context(FeelContext),
@@ -85,8 +100,14 @@ pub enum Value {
   /// Value representing a key of the context entry.
   ContextEntryKey(Name),
 
-  /// Value representing a name of the context entry in context type definition.
-  ContextEntryName(Name),
+  /// Value representing the context type.
+  ContextType(FeelType),
+
+  /// Value representing a context entry in context type definition.
+  ContextTypeEntry(Name, FeelType),
+
+  /// Value representing a key of the context entry in context type definition.
+  ContextTypeEntryKey(Name),
 
   /// Value for storing dates as [FeelDate].
   Date(FeelDate),
@@ -97,6 +118,9 @@ pub enum Value {
   /// Value for days and time durations.
   DaysAndTimeDuration(FeelDaysAndTimeDuration),
 
+  /// Value representing the `FEEL` type of a value.
+  FeelType(FeelType),
+
   /// Value representing function's formal parameter with name and type.
   FormalParameter(Name, FeelType),
 
@@ -107,8 +131,8 @@ pub enum Value {
   FunctionBody(FunctionBody),
 
   /// Value representing the function definition.
-  /// This value holds the list of function's formal parameters, and the function's body.
-  FunctionDefinition(Vec<(Name, FeelType)>, FunctionBody),
+  /// This value holds the list of function's formal parameters, the function's body and expected result type.
+  FunctionDefinition(Vec<(Name, FeelType)>, FunctionBody, FeelType),
 
   /// Value representing interval end.
   IntervalEnd(Box<Value>, bool),
@@ -116,7 +140,7 @@ pub enum Value {
   /// Value representing interval start.
   IntervalStart(Box<Value>, bool),
 
-  /// **Irrelevant** value...
+  /// Value representing `FEEL` `irrelevant` value.
   Irrelevant,
 
   /// Value representing a list of values.
@@ -125,23 +149,29 @@ pub enum Value {
   /// Named parameter.
   NamedParameter(Box<Value>, Box<Value>),
 
-  /// Value representing a collection of name parameters.
-  NamedParameters(BTreeMap<Name, Value>),
+  /// Value representing a collection of named parameters.
+  NamedParameters(BTreeMap<Name, (Value, usize)>),
 
-  /// **NegatedCommaList** value...
-  NegatedCommaList(Vec<Value>),
+  /// Value representing a collection of values representing a negated comma-separated list of expressions.
+  NegatedCommaList(Values),
 
   /// Null value with optional tracing message.
   Null(Option<String>),
 
-  /// Numeric value
-  Number(f64),
+  /// Value representing `FEEL` number type.
+  Number(FeelNumber),
 
   /// Name of the parameter.
   ParameterName(Name),
 
+  /// Value representing a list of function's parameter types.
+  ParameterTypes(Vec<Value>),
+
   /// List of positional parameters.
-  PositionalParameters(Vec<Value>),
+  PositionalParameters(Values),
+
+  /// Value representing a segment of a qualified name.
+  QualifiedNameSegment(Name),
 
   /// Value representing a `range`.
   Range(Box<Value>, bool, Box<Value>, bool),
@@ -152,9 +182,6 @@ pub enum Value {
   /// Value for storing time as [FeelTime].
   Time(FeelTime),
 
-  /// Value representing only the `FEEL` type of value.
-  Type(FeelType),
-
   /// **UnaryGreater** value...
   UnaryGreater(Box<Value>),
 
@@ -171,52 +198,59 @@ pub enum Value {
   YearsAndMonthsDuration(FeelYearsAndMonthsDuration),
 }
 
-impl Stringify for Value {
-  /// Converts a [Value] to its `TEXT` representation.
-  fn stringify(&self) -> String {
-    self.feelify()
+impl std::fmt::Display for Value {
+  fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+    match self {
+      Value::Boolean(value) => write!(f, "{}", value),
+      Value::BuiltInFunction(_) => write!(f, "BuiltInFunction"),
+      Value::ExpressionList(items) => write!(f, "{}", items),
+      Value::Context(context) => write!(f, "{}", context),
+      Value::ContextEntry(_, _) => write!(f, "ContextEntry"),
+      Value::ContextEntryKey(name) => write!(f, "{}", name),
+      Value::ContextType(_) => write!(f, "ContextType"),
+      Value::ContextTypeEntry(name, feel_type) => write!(f, "{}: {}", name, feel_type),
+      Value::ContextTypeEntryKey(name) => write!(f, "{}", name),
+      Value::Date(date) => write!(f, "{}", date),
+      Value::DateTime(date_time) => write!(f, "{}", date_time),
+      Value::DaysAndTimeDuration(dt_duration) => write!(f, "{}", dt_duration),
+      Value::FeelType(feel_type) => write!(f, "type({})", feel_type),
+      Value::FormalParameter(_, _) => write!(f, "FormalParameter"),
+      Value::FormalParameters(_) => write!(f, "FormalParameters"),
+      Value::FunctionBody(_) => write!(f, "FunctionBody"),
+      Value::FunctionDefinition { .. } => write!(f, "FunctionDefinition"),
+      Value::IntervalEnd(_, _) => write!(f, "IntervalEnd"),
+      Value::IntervalStart(_, _) => write!(f, "IntervalStart"),
+      Value::Irrelevant => write!(f, "Irrelevant"),
+      Value::List(items) => write!(f, "{}", items),
+      Value::NamedParameter(_, _) => write!(f, "NamedParameter"),
+      Value::NamedParameters(_) => write!(f, "NamedParameters"),
+      Value::NegatedCommaList(_) => write!(f, "NegatedCommaList"),
+      Value::Number(value) => write!(f, "{}", value),
+      Value::Null(trace) => write!(f, "null{}", trace.as_ref().map_or("".to_string(), |s| format!("({})", s))),
+      Value::ParameterName(_) => write!(f, "ParameterName"),
+      Value::ParameterTypes(_) => write!(f, "ParameterTypes"),
+      Value::PositionalParameters(_) => write!(f, "PositionalParameters"),
+      Value::QualifiedNameSegment(_) => write!(f, "QualifiedNameSegment"),
+      Value::Range(_, _, _, _) => write!(f, "Range"),
+      Value::String(s) => write!(f, "\"{}\"", s),
+      Value::Time(time) => write!(f, "{}", time),
+      Value::UnaryGreater(_) => write!(f, "UnaryGreater"),
+      Value::UnaryGreaterOrEqual(_) => write!(f, "UnaryGreaterOrEqual"),
+      Value::UnaryLess(_) => write!(f, "UnaryLess"),
+      Value::UnaryLessOrEqual(_) => write!(f, "UnaryLessOrEqual"),
+      Value::YearsAndMonthsDuration(ym_duration) => write!(f, "{}", ym_duration),
+    }
   }
 }
 
-impl Feelify for Value {
-  /// Converts a [Value] to its `FEEL` representation.
-  fn feelify(&self) -> String {
+impl ToFeelString for Value {
+  /// Converts [Value] into `FEEL` string.
+  fn to_feel_string(&self) -> String {
     match self {
-      Value::Boolean(b) => format!("{}", b),
-      Value::BuiltInFunction(_) => "BuiltInFunction".to_string(),
-      Value::BuiltInType(_) => "BuiltInType".to_string(),
-      Value::CommaList(items) => items.feelify(),
-      Value::Context(ctx) => ctx.feelify(),
-      Value::ContextEntry(_, _) => "ContextEntry".to_string(),
-      Value::ContextEntryKey(name) => name.feelify(),
-      Value::ContextEntryName(name) => name.feelify(),
-      Value::Date(date) => format!("{}", date),
-      Value::DateTime(date_time) => format!("{}", date_time),
-      Value::DaysAndTimeDuration(duration) => duration.clone().into(),
-      Value::FormalParameter(_, _) => "FormalParameter".to_string(),
-      Value::FormalParameters(_) => "FormalParameters".to_string(),
-      Value::FunctionBody(_) => "FunctionBody".to_string(),
-      Value::FunctionDefinition(_, _) => "FunctionDefinition".to_string(),
-      Value::IntervalEnd(_, _) => "IntervalEnd".to_string(),
-      Value::IntervalStart(_, _) => "IntervalStart".to_string(),
-      Value::Irrelevant => "Irrelevant".to_string(),
-      Value::List(items) => items.feelify(),
-      Value::NamedParameter(_, _) => "NamedParameter".to_string(),
-      Value::NamedParameters(_) => "NamedParameters".to_string(),
-      Value::NegatedCommaList(_) => "NegatedCommaList".to_string(),
-      Value::Number(n) => format!("{}", n),
-      Value::Null(trace) => format!("null{}", trace.as_ref().map_or("".to_string(), |s| format!(" ({})", s))),
-      Value::ParameterName(_) => "ParameterName".to_string(),
-      Value::PositionalParameters(_) => "PositionalParameters".to_string(),
-      Value::Range(_, _, _, _) => "Range".to_string(),
-      Value::String(s) => format!(r#""{}""#, s),
-      Value::Time(time) => format!("{}", time),
-      Value::Type(feel_type) => format!("type({:?})", feel_type),
-      Value::UnaryGreater(_) => "UnaryGreater".to_string(),
-      Value::UnaryGreaterOrEqual(_) => "UnaryGreaterOrEqual".to_string(),
-      Value::UnaryLess(_) => "".to_string(),
-      Value::UnaryLessOrEqual(_) => "UnaryLessOrEqual".to_string(),
-      Value::YearsAndMonthsDuration(duration) => duration.clone().into(),
+      Value::Context(context) => context.to_feel_string(),
+      Value::List(items) => items.to_feel_string(),
+      Value::String(value) => format!("\"{}\"", value.replace("\"", "\\\"")),
+      other => other.to_string(),
     }
   }
 }
@@ -225,15 +259,15 @@ impl Jsonify for Value {
   /// Converts a [Value] to its `JSON` representation.
   fn jsonify(&self) -> String {
     match self {
-      Value::Boolean(b) => format!("{}", b),
-      Value::CommaList(items) => items.feelify(),
+      Value::Boolean(value) => format!("{}", value),
+      Value::ExpressionList(items) => items.to_string(),
       Value::Context(ctx) => ctx.jsonify(),
-      Value::ContextEntryKey(name) => name.feelify(),
+      Value::ContextEntryKey(name) => name.to_string(),
       Value::List(items) => items.jsonify(),
-      Value::Number(n) => format!("{}", n),
+      Value::Number(value) => value.jsonify(),
       Value::Null(_) => "null".to_string(),
       Value::String(s) => format!("\"{}\"", s),
-      _ => format!("jsonify not implemented for: {:?}", self),
+      _ => format!("jsonify not implemented for: {}", self),
     }
   }
 }
@@ -256,8 +290,7 @@ impl Value {
     match self {
       Value::Boolean(_) => FeelType::Boolean,
       Value::BuiltInFunction(_) => FeelType::Any,
-      Value::BuiltInType(feel_type) => feel_type.clone(),
-      Value::CommaList(_) => FeelType::Any,
+      Value::ExpressionList(_) => FeelType::Any,
       Value::Context(context) => {
         let mut entries = BTreeMap::new();
         for (name, value) in context.deref() {
@@ -267,26 +300,29 @@ impl Value {
       }
       Value::ContextEntry(_, _) => FeelType::Any,
       Value::ContextEntryKey(_) => FeelType::Any,
-      Value::ContextEntryName(_) => FeelType::Any,
+      Value::ContextType(feel_type) => feel_type.clone(),
+      Value::ContextTypeEntry(_, feel_type) => feel_type.clone(),
+      Value::ContextTypeEntryKey(_) => FeelType::Any,
       Value::Date(_) => FeelType::Date,
       Value::DateTime(_) => FeelType::DateTime,
       Value::DaysAndTimeDuration(_) => FeelType::DaysAndTimeDuration,
+      Value::FeelType(feel_type) => feel_type.clone(),
       Value::FormalParameter(_, feel_type) => feel_type.clone(),
       Value::FormalParameters(_) => FeelType::Any,
       Value::FunctionBody(_) => FeelType::Any,
-      Value::FunctionDefinition(parameters, _) => {
+      Value::FunctionDefinition(parameters, _, result_type) => {
         let parameter_types = parameters.iter().map(|(_, feel_type)| feel_type.clone()).collect();
-        FeelType::Function(parameter_types, Box::new(FeelType::Any))
+        FeelType::Function(parameter_types, Box::new(result_type.clone()))
       }
       Value::IntervalEnd(interval_end, _) => interval_end.type_of(),
       Value::IntervalStart(interval_start, _) => interval_start.type_of(),
       Value::Irrelevant => FeelType::Any,
       Value::List(values) => {
-        if values.to_vec().is_empty() {
+        if values.as_vec().is_empty() {
           FeelType::List(Box::new(FeelType::Null))
         } else {
-          let item_type = values.to_vec()[0].type_of();
-          for item in values.to_vec() {
+          let item_type = values.as_vec()[0].type_of();
+          for item in values.as_vec() {
             if item.type_of() != item_type {
               return FeelType::List(Box::new(FeelType::Any));
             }
@@ -300,7 +336,9 @@ impl Value {
       Value::Null(_) => FeelType::Null,
       Value::Number(_) => FeelType::Number,
       Value::ParameterName(_) => FeelType::Any,
+      Value::ParameterTypes(_) => FeelType::Any,
       Value::PositionalParameters(_) => FeelType::Any,
+      Value::QualifiedNameSegment(_) => FeelType::Any,
       Value::Range(range_start, _, range_end, _) => {
         let range_start_type = range_start.type_of();
         let range_end_type = range_end.type_of();
@@ -311,7 +349,6 @@ impl Value {
       }
       Value::String(_) => FeelType::String,
       Value::Time(_) => FeelType::Time,
-      Value::Type(feel_type) => feel_type.clone(),
       Value::UnaryGreater(_) => FeelType::Boolean,
       Value::UnaryGreaterOrEqual(_) => FeelType::Boolean,
       Value::UnaryLess(_) => FeelType::Boolean,
@@ -319,35 +356,19 @@ impl Value {
       Value::YearsAndMonthsDuration(_) => FeelType::YearsAndMonthsDuration,
     }
   }
-  /// Applies coercion rules to this value and returns converted value or `null`.
-  pub fn coerce(&self, feel_type: &FeelType) -> Value {
-    // coercion of singleton list into single value when single type expected
-    if let Value::List(items) = self {
-      if items.to_vec().len() == 1 && items.to_vec()[0].type_of().is_conformant(feel_type) {
-        return items.to_vec()[0].clone();
-      }
-    }
-    // coercion of single value into singleton list when expected a list
-    if let FeelType::List(inner_feel_type) = feel_type {
-      if self.type_of().is_conformant(inner_feel_type) {
-        return Value::List(Values::new(vec![self.clone()]));
-      }
-    }
-    null_with_trace!("coercion of value `{}` into type `{:?}` results with `null`", self.feelify(), feel_type)
-  }
   /// Tries to convert `xsd:integer` string into valid [Value] representing a number.
   pub fn try_from_xsd_integer(text: &str) -> Result<Self> {
-    let value = text.parse::<i64>().map_err(|_| invalid_xsd_integer(text))?;
-    Ok(Value::Number(value as f64))
+    let value = text.parse::<FeelNumber>().map_err(|_| invalid_xsd_integer(text))?;
+    Ok(Value::Number(value))
   }
   /// Tries to convert `xsd:decimal` string into valid [Value] representing a number.
   pub fn try_from_xsd_decimal(text: &str) -> Result<Self> {
-    let value = text.parse::<f64>().map_err(|_| invalid_xsd_decimal(text))?;
+    let value = text.parse::<FeelNumber>().map_err(|_| invalid_xsd_decimal(text))?;
     Ok(Value::Number(value))
   }
   /// Tries to convert `xsd:double` string into valid [Value] representing a number.
   pub fn try_from_xsd_double(text: &str) -> Result<Self> {
-    let value = text.parse::<f64>().map_err(|_| invalid_xsd_double(text))?;
+    let value = text.parse::<FeelNumber>().map_err(|_| invalid_xsd_double(text))?;
     Ok(Value::Number(value))
   }
   /// Tries to convert `xsd:boolean` string into valid [Value] representing a boolean.
@@ -393,57 +414,68 @@ impl Value {
 }
 
 /// Collection of values.
-#[derive(Debug, Clone, PartialEq)]
+#[derive(Debug, Clone, PartialEq, Default)]
 pub struct Values(Vec<Value>);
 
 impl Values {
-  pub fn empty() -> Self {
-    Self(vec![])
-  }
+  ///
   pub fn new(values: Vec<Value>) -> Self {
     Self(values)
   }
+  ///
   pub fn add(&mut self, value: Value) {
     self.0.push(value);
   }
+  ///
   pub fn insert(&mut self, index: usize, value: Value) {
     self.0.insert(index, value);
   }
+  ///
   pub fn remove(&mut self, index: usize) {
     self.0.remove(index);
   }
+  ///
   pub fn len(&self) -> usize {
     self.0.len()
   }
+  ///
   pub fn is_empty(&self) -> bool {
     self.0.is_empty()
   }
+  ///
   pub fn reverse(&mut self) {
     self.0.reverse();
   }
-  pub fn to_vec(&self) -> &Vec<Value> {
+  ///
+  pub fn as_vec(&self) -> &Vec<Value> {
     &self.0
   }
 }
 
-impl Feelify for Values {
-  ///
-  fn feelify(&self) -> String {
-    format!("[{}]", self.0.iter().map(|value| value.feelify()).collect::<Vec<String>>().join(","))
+impl std::fmt::Display for Values {
+  fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+    write!(f, "[{}]", self.0.iter().map(|value| value.to_string()).collect::<Vec<String>>().join(", "))
+  }
+}
+
+impl ToFeelString for Values {
+  /// Converts [Values] into `FEEL` string.
+  fn to_feel_string(&self) -> String {
+    format!("[{}]", self.0.iter().map(|value| value.to_feel_string()).collect::<Vec<String>>().join(", "))
   }
 }
 
 impl Jsonify for Values {
   ///
   fn jsonify(&self) -> String {
-    format!("[{}]", self.0.iter().map(|value| value.jsonify()).collect::<Vec<String>>().join(","))
+    format!("[{}]", self.0.iter().map(|value| value.jsonify()).collect::<Vec<String>>().join(", "))
   }
 }
 
 /// Definitions of value errors.
 pub mod errors {
   use crate::values::Value;
-  use dmntk_common::{DmntkError, Feelify};
+  use dmntk_common::DmntkError;
 
   /// Value errors.
   #[derive(Debug, PartialEq)]
@@ -532,39 +564,76 @@ pub mod errors {
   }
 
   pub fn value_is_not_a_context(value: &Value) -> DmntkError {
-    ValueError::ValueIsNotAContext(value.feelify()).into()
+    ValueError::ValueIsNotAContext(value.to_string()).into()
   }
 }
 
 #[cfg(test)]
 mod tests {
   use crate::context::FeelContext;
+  use crate::function::FunctionBody;
   use crate::values::Value;
+  use crate::{FeelNumber, Scope, ToFeelString};
+  use std::rc::Rc;
 
   #[test]
-  fn test_value_is_number() {
-    assert!(Value::Number(10.).is_number());
-    assert!(Value::Number(10.123).is_number());
+  fn _0001() {
+    assert!(value_number!(10).is_number());
+    assert!(value_number!(10123, 3).is_number());
     assert!(!Value::Boolean(true).is_number());
     assert!(!Value::Boolean(false).is_number());
   }
 
   #[test]
-  fn test_value_equality_number() {
-    assert!((Value::Number(10.) == Value::Number(10.)));
-    assert!((Value::Number(10.) != Value::Number(10.1)));
+  fn _0002() {
+    assert_eq!(value_number!(10), value_number!(10));
+    assert_ne!(value_number!(10, 0), value_number!(10, 1));
   }
 
   #[test]
-  fn test_value_equality_boolean() {
-    assert!((Value::Boolean(true) == Value::Boolean(true)));
-    assert!((Value::Boolean(false) == Value::Boolean(false)));
-    assert!((Value::Boolean(true) != Value::Boolean(false)));
-    assert!((Value::Boolean(false) != Value::Boolean(true)));
+  fn _0003() {
+    assert_eq!(Value::Boolean(true), Value::Boolean(true));
+    assert_eq!(Value::Boolean(false), Value::Boolean(false));
+    assert_ne!(Value::Boolean(true), Value::Boolean(false));
+    assert_ne!(Value::Boolean(false), Value::Boolean(true));
   }
 
   #[test]
-  fn test_value_equality_context() {
-    assert!((Value::Context(FeelContext::default()) == Value::Context(FeelContext::default())));
+  fn _0004() {
+    assert_eq!(Value::Context(FeelContext::default()), Value::Context(FeelContext::default()));
+  }
+
+  #[test]
+  fn _0005() {
+    let fun_body_a = FunctionBody::Context(Rc::new(Box::new(|_: &Scope| value_number!(1))));
+    let fun_body_b = FunctionBody::Context(Rc::new(Box::new(|_: &Scope| value_number!(2))));
+    let fun_body_c = FunctionBody::LiteralExpression(Rc::new(Box::new(|_: &Scope| value_number!(3))));
+    let fun_body_d = FunctionBody::LiteralExpression(Rc::new(Box::new(|_: &Scope| value_number!(4))));
+    let fun_body_e = FunctionBody::DecisionTable(Rc::new(Box::new(|_: &Scope| value_number!(4))));
+    let fun_body_f = FunctionBody::DecisionTable(Rc::new(Box::new(|_: &Scope| value_number!(5))));
+    let fun_body_g = FunctionBody::External(Rc::new(Box::new(|_: &Scope| value_number!(6))));
+    let fun_body_h = FunctionBody::External(Rc::new(Box::new(|_: &Scope| value_number!(7))));
+    assert_eq!(fun_body_a, fun_body_b);
+    assert_ne!(fun_body_a, fun_body_c);
+    assert_ne!(fun_body_a, fun_body_e);
+    assert_ne!(fun_body_a, fun_body_g);
+    assert_eq!(fun_body_c, fun_body_d);
+    assert_ne!(fun_body_c, fun_body_a);
+    assert_ne!(fun_body_c, fun_body_e);
+    assert_ne!(fun_body_c, fun_body_g);
+    assert_eq!(fun_body_e, fun_body_f);
+    assert_ne!(fun_body_e, fun_body_a);
+    assert_ne!(fun_body_e, fun_body_c);
+    assert_ne!(fun_body_e, fun_body_g);
+    assert_eq!(fun_body_g, fun_body_h);
+    assert_ne!(fun_body_g, fun_body_a);
+    assert_ne!(fun_body_g, fun_body_c);
+    assert_ne!(fun_body_g, fun_body_e);
+  }
+
+  #[test]
+  fn test_value_to_feel_string() {
+    assert_eq!(r#""foo""#, Value::String("foo".to_string()).to_feel_string());
+    assert_eq!(r#""\"bar\"""#, Value::String("\"bar\"".to_string()).to_feel_string());
   }
 }
